
#include "terranova.h"

#define Parameter(type, name, defval) \
    extern type name;

#include "parameters.h"

static int seeded = FALSE;

static const struct option long_options[] = {
    { "blackhole", 0, NULL, 221 },
    { "clouds", 0, NULL, 202 },
    { "clusterdensity", 1, NULL, 214 },
    { "clusters", 1, NULL, 213 },
    { "clusterwidth", 1, NULL, 215 },
    { "copyright", 0, NULL, 'c' },
    { "crescents", 0, NULL, 217 },
    { "dimclouds", 1, NULL, 206 },
    { "dimplanet", 1, NULL, 205 },
    { "glaciers", 1, NULL, 210 },
    { "height", 1, NULL, 'h' },
    { "hell", 0, NULL, 216 },
    { "help", 0, NULL, 'u' },
    { "ice", 1, NULL, 209 },
    { "mesh", 1, NULL, 'm' },
    { "planet", 0, NULL, 203 },
    { "powerclouds", 1, NULL, 208 },
    { "powerplanet", 1, NULL, 207 },
    { "random", 0, NULL, 'r' },
    { "seed", 1, NULL, 's' },
    { "starbright", 1, NULL, 219 },
    { "starcolour", 1, NULL, 212 },
    { "starcolor", 1, NULL, 212 },
    { "stardensity", 1, NULL, 211 },
    { "stargamma", 1, NULL, 220 },
    { "stars", 0, NULL, 204 },
    { "starsize", 1, NULL, 218 },
    { "verbose", 0, NULL, 'v' },
    { "version", 0, NULL, 201 },
    { "width", 1, NULL, 'w' },
    { 0, 0, 0, 0 }
};
static const char single_letter_options[] = "ch:m:rs:vuw:";
static int option_index = 0;

static void usage(void)
{
    printf("Usage:  terranova\n");
}

/*  SETRANGE_DOUBLE  --  Set a floating point range parameter.
    SETRANGE_INT     --  Set an integer range parameter.  */

#define setrange(type, scan, format) \
static void setrange_##type(type *pmin, type *pmax, const type umin, const type umax) \
{ \
    char *sep = strchr(optarg, ','); \
    if (sep != NULL) { \
    	*sep = 0; \
	*pmin = scan(optarg); \
	*pmax = scan(sep + 1); \
    } else { \
    	*pmin = *pmax = scan(optarg); \
    } \
    if (*pmin > *pmax) { \
    	type a = *pmin; \
	*pmin = *pmax; \
	*pmax = a; \
    } \
    *pmin = min(max(*pmin, umin), umax); \
    *pmax = min(max(*pmax, umin), umax); \
/* fprintf(stderr, "Min = " format "  Max = " format "\n", *pmin, *pmax); */ \
}

setrange(int, atoi, "%d")
setrange(double, atof, "%.4f")

#define rangeopt(type, name) setrange_##type(&name##_min, &name##_max, name##_umin, name##_umax)

int main(int argc, char *argv[])
{
    int opt;
    
    while ((opt = getopt_long(argc, argv, single_letter_options,
    	    	    long_options, &option_index)) != -1) {

        switch (opt) {
	
	    case 221:	    	    /* --blackhole */
	    	blackhole = TRUE;
		goto planet;
	
	    case 202:	    	    /* --clouds */
heaven:     	show_clouds = TRUE;
		show_planets = show_stars = FALSE;
		break;
		
	    case 214:	    	    /* --clusterdensity min[,max]  Star cluster density */
	    	rangeopt(int, star_cluster_density);
		break;
		
	    case 213:	    	    /* --clusters min[,max]  Number of star clusters */
	    	rangeopt(int, star_clusters);
		break;
		
	    case 215:	    	    /* --clusterwidth min[,max]  Star cluster width */
	    	rangeopt(int, star_cluster_width);
		break;

            case 'c':	    	    /* --copyright */
                printf("This program is in the public domain.\n");
                return 0;
		
	    case 217:	    	    /* --crescents  Permit extreme crescent planets */
	    	crescents = TRUE;
		break;
		
	    case 206:	    	    /* --dimclouds min[,max]  Fractal dimension of clouds */
	    	rangeopt(double, fractal_dimension_clouds);
		break;
		
	    case 205:	    	    /* --dimplanet min[,max]  Fractal dimension of planet */
	    	rangeopt(double, fractal_dimension);
		break;
		
	    case 210:	    	    /* --glaciers n   Glacier level */
	    	rangeopt(double, glacier_level);
		break;
		
	    case 'h':	    	    /* -h --height n 	Image height n pixels */
	    	screenY = atoi(optarg);
		break;
		
	    case 216:	    	    /* --hell  Show Hellfire */
	    	hell = TRUE;
		goto heaven;
		
	    case 209:	    	    /* --ice n   Ice level */
	    	rangeopt(double, ice_level);
		break;
		
	    case 'm':	    	    /* -m --mesh n 	Mesh size n by n */
	    	mesh_size = atoi(optarg);
		break;
		
	    case 203:	    	    /* --planet */
planet:	    	show_planets = TRUE;
		show_clouds = show_stars = FALSE;
		break;
		
	    case 208:	    	    /* --powerclouds min[,max]  Power spectrum of clouds */
	    	rangeopt(double, power_spectrum_clouds);
		break;
		
	    case 207:	    	    /* --powerplanet min[,max]  Power spectrum of planet */
	    	rangeopt(double, power_spectrum);
		break;
		
            case 'r':	    	    /* -r  --random  Randomise planet position */
	    	randomPosition = TRUE;
                break;
		
	    case 's':	    	    /* -s --seed n 	Seed pseudorandom generator with n */
	    	srandom(atoi(optarg));
		seeded = TRUE;
		break;
		
	    case 219:	    	  /*  --starbright f  Adjust star intensity by factor f  */
		star_bright = atof(optarg);
		break;
		
	    case 212:	    	    /* --starcolour min[,max]  Star colour saturation */
	    	rangeopt(int, star_colour);
		break;
		
	    case 211:	    	    /* --stardensity min[,max]  Star density */
	    	rangeopt(int, star_fraction);
		break;
		
	    case 220:	    	  /*  --stargamma f  Gamma exponent for star brightness  */
		star_gamma = atof(optarg);
		break;
		
	    case 204:	    	    /* --stars */
	    	show_stars = TRUE;
		show_planets = show_clouds = FALSE;
		break;
		
	    case 218:	    	  /*  --starsize f  Width of Gaussian point spread function for stars.
	    	    	    	    	    	    This is a floating point value; smaller values
						    produce larger stars.  */
		star_size = atof(optarg);
		break;
		
            case 'u':             /* -u --help  Print how-to-call information*/
            case '?':
                usage();
                return 0;
		
	    case 'v':	    	  /* --verbose */
	    	verbose = TRUE;
		break;
		
            case 201:             /* --version  */
            	printf("Version 2.1: January 2006\n");
            	return 0;
		
	    case 'w':	    	    /* -w --width n 	Image width n pixels */
	    	screenX = atoi(optarg);
		break;
		
	    default:
	    	fprintf(stderr, "***Internal error: unhandled case %d in option processing.\n",
	    	    opt);
		return 1;
        }
    }
    
    if (!seeded) {
    	unsigned int s = time(NULL) ^ getpid() ^ (getppid() << 3);
	int i, j;
	
	
	srandom(s);
	
	s = random();
	
	s ^= random() << 9;
	
	srandom(s);
	
	j = random();
	
	for (i = 0; i < (j & 31); i++) {
	    (void) random();
	}
	
    	srandom(s);
	if (verbose) {
	    fprintf(stderr, "Pseudorandom seed: %u\n", s);
	}
    }

    create_ppm(stdout);

    return 0;
}
